/**
 * Copyright (C) 2003-2004 Funambol
 *
 *  This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 */
package sync4j.framework.server.session;

import java.util.*;

import sync4j.framework.core.*;
import sync4j.framework.core.Map;
import sync4j.framework.security.Sync4jPrincipal;
import sync4j.framework.security.SecurityConstants;
import sync4j.framework.server.Sync4jDevice;
import sync4j.framework.protocol.ProtocolUtil;

/** 
 * This class represents the state of the Syncronization process because cache 
 * all informations that could be used in that process.
 * It is set during the Initialization process.
 *
 * @author Luigia Fassina @ funambol.com
 * @version $Id: SyncState.java,v 1.8 2004/07/07 14:54:12 luigiafassina Exp $
 */
public class SyncState
implements SecurityConstants {
    
    // ------------------------------------------------------------- Public data

    /**
     * Cache the Sync4jDevice object (it contains Sync4j information and nonces
     */
    public Sync4jDevice device;
    
    /**
     * Authenticated credential. If null no credential has authenticated.
     */
    public Cred loggedCredential;
    
    /**
     * Authenticated principal. If null, no principal has authenticated
     */
    public Sync4jPrincipal loggedPrincipal;
    
    /**
     * Authentication state. One of the following values:
     * <ul>
     *   <li>AUTH_UNAUTHENTICATED</li>
     *   <li>AUTH_MISSING_CREDENTIALS</li>
     *   <li>AUTH_INVALID_CREDENTIALS</li>
     *   <li>AUTH_AUTHENTICATED</li>
     * </ul>
     */
    public int authenticationState;
    
    /**
     * Server authentication state. One of the following values:
     * <ul>
     *   <li>AUTH_UNAUTHENTICATED</li>
     *   <li>AUTH_RETRY_1</li>
     *   <li>AUTH_RETRY_2</li>
     * </ul>
     */
    public int serverAuthenticationState;
    
    /**
     * Has the session been started ?
     */
    public boolean started;
    
    /**
     * Cache the SyncML Protocol version used by device
     */
    public String syncMLVerProto;
    
    // ---------------------------------------------------------- Public Methods
    public SyncState() {
        reset();
    }
    
    /**
     * Reset this object to an initial state
     */
    public void reset() {
        device                    = null;
        loggedCredential          = null;
        loggedPrincipal           = null;
        authenticationState       = AUTH_UNAUTHENTICATED;
        serverAuthenticationState = AUTH_UNAUTHENTICATED;
        started                   = false;
        syncMLVerProto            = null;
    }
    
    /**
     * Cache all AbstractCommand (Alert, Put, Get)
     */
    private ArrayList listClientCommands = new ArrayList();
   
    public void addClientCommands(AbstractCommand[] clientCommands) {
        for (int i=0; ((clientCommands != null) && (i<clientCommands.length)); i++) {
            listClientCommands.add(clientCommands[i]);
        }
    }
    
    public AbstractCommand[] getClientCommands() {
        return (AbstractCommand[])listClientCommands.toArray(new AbstractCommand[listClientCommands.size()]);
    }

    /**
     * Cache the AlertCommand
     */
    private ArrayList listClientAlerts = new ArrayList();
   
    public void addClientAlerts(Alert[] clientAlerts) {
        for (int i=0; ((clientAlerts != null) && (i<clientAlerts.length)); i++) {
            listClientAlerts.add(clientAlerts[i]);
        }
    }
    
    public void resetClientAlerts() {
        listClientAlerts = new ArrayList();
    }
    
    public void removeClientAlert(Alert alert) {
        listClientAlerts.remove(alert);
    }
    
    public Alert[] getClientAlerts() {
        return (Alert[])listClientAlerts.toArray(new Alert[listClientAlerts.size()]);
    }
    
    /**
     * Cache the MapCommand
     */
    private ArrayList listMapCommands = new ArrayList();
    
    public void addMapCommands(Map[] mapCommands) {
        for (int i=0; ((mapCommands != null) && (i<mapCommands.length)); i++) {
            listMapCommands.add(mapCommands[i]);
        }
    }
    
    public Map[] getMapCommands() {
        return (Map[])listMapCommands.toArray(new Map[listMapCommands.size()]);
    }
    
    /**
     * Cache the SyncCommand
     */
    private ArrayList listSyncCommands = new ArrayList();
    
    public void addSyncCommands(Sync[] syncCommands) {
        for (int i=0; ((syncCommands != null) && (i<syncCommands.length)); i++) {
            listSyncCommands.add(syncCommands[i]);
        }
    }
    
    public Sync[] getSyncCommands() {
        return (Sync[])listSyncCommands.toArray(new Sync[listSyncCommands.size()]);
    }
    
    /**
     * Cache the server modification commands: in the case of multimessage these 
     * commands must be sent only when the client has sent its package
     */
    private ArrayList listServerModifications = new ArrayList();
    
    public void setServerModifications(AbstractCommand[] serverModifications) {
        for (int i=0; ((serverModifications != null) && (i<serverModifications.length)); i++) {
            listServerModifications.add(serverModifications[i]);
        }
    }
    
    public AbstractCommand[] getServerModifications() {
        return (AbstractCommand[])listServerModifications.toArray(new AbstractCommand[listServerModifications.size()]);
    }
    
    /**
     * Cache the Status command presents in the response message
     */
    private LinkedList listStatusCmdOut = new LinkedList();
    
    public void addStatusCmdOut(LinkedList statusList) {    
        listStatusCmdOut.addAll(statusList);
    }
    
    public LinkedList getStatusCmdOut() {
        return listStatusCmdOut;
    }
    
    public void removeStatusCmdOut(ArrayList statusCommand) {
        listStatusCmdOut.removeAll(statusCommand);
    }
    
    /**
     * Cache the Status command for Map presents in the response message
     */
    private LinkedList listMapStatusOut = new LinkedList();
    
    public void addMapStatusOut(LinkedList mapStatusList) {    
        listMapStatusOut.addAll(mapStatusList);
    }
    
    public LinkedList getMapStatusOut() {
        return listMapStatusOut;
    }
    
    public void removeMapStatusOut(List mapStatus) {
        listMapStatusOut.removeAll(mapStatus);
    }
    
    /**
     * Cache the Alert command presents in the response message
     */
    private LinkedList listAlertCmdOut = new LinkedList();
    
    public void addAlertCmdOut(LinkedList alertList) {
        listAlertCmdOut.addAll(alertList);
    }
    
    public LinkedList getAlertCmdOut() {
        return listAlertCmdOut;
    }
    
    public void removeAlertCmdOut(ArrayList alertList) {
        listAlertCmdOut.removeAll(alertList);
    }
    
    /**
     * Cache the AbstractCommand presents in the response message
     */
    private LinkedList listCmdOut = new LinkedList();
    
    public void addCmdOut(LinkedList cmdList) {
        listCmdOut.addAll(cmdList);
    }
    
    public LinkedList getCmdOut() {
        return listCmdOut;
    }
    
    public void removeCmdOut(ArrayList abstractCommand) {
        listCmdOut.removeAll(abstractCommand);
    }
   
    /**
     * Cache the maximum message size
     */
    private long maxMsgSize = 0;
    public void setMaxMsgSize(long value) {
        maxMsgSize = value;
    }
    
    public long getMaxMsgSize() {
        return maxMsgSize;
    }
    
    /**
     * Is Sync with initialization?
     */
    private boolean syncWithInit = false;
    public void setSyncWithInit(boolean syncWithInit) {
        this.syncWithInit = syncWithInit;
    }
    public boolean isSyncWithInit() {
        return this.syncWithInit;
    }
    
    /**
     * Cache the response with initialization
     */
    private SyncML responseInit = null;
    public void setResponseInit(SyncML responseInit) {
        this.responseInit = responseInit;
    }
    public SyncML getResponseInit() {
        return this.responseInit;
    }
    
    /**
     * Is the final response?
     */
    private boolean respFinal = true;
    public void setResponseFinal(boolean respFinal) {
        this.respFinal = respFinal;
    }
    public boolean isResponseFinal() {
        return this.respFinal;
    }

    /**
     * Cache overhead for SyncHdr object
     */
    private long overheadHdr = 0;
    public long getOverheadHdr() {
        return this.overheadHdr;
    }
    public void setOverheadHdr(long overheadHdr) {
        this.overheadHdr = overheadHdr;
    }

    /**
     * Cache overhead of Status for SyncHdr
     */
    private long sizeStatusSyncHdr = 0;
    public long getSizeStatusSyncHdr() {
        return this.sizeStatusSyncHdr;
    }
    public void setSizeStatusSyncHdr(long sizeStatusSyncHdr) {
        this.sizeStatusSyncHdr = sizeStatusSyncHdr;
    }
    
    /**
     * Cache Status for SyncHdr
     */
    private Status statusSyncHdr;
    public void setStatusSyncHdr(Status statusSyncHdr) {
        this.statusSyncHdr = statusSyncHdr;
    }
    public Status getStatusSyncHdr() {
        return this.statusSyncHdr;
    }
}
